#! /usr/bin env python

"""
Code to generate NIRCAM dispersed seed images. Starting with a set of imaging seed images, 
potentially with padding, given a NIRCAM GRISMCONF configuration file
"""

import os
from astropy.io import fits
import numpy as np
from .observations.observations \
        import observation as NIRCAM_Gsim_observation 

class Grism_seed():
	def __init__(self,image_seeds,cross_filter,mode,config_path=".",extrapolate_SED=False,SED_file=None):
		# image_seeds: "V4*.fits"
		# mode: "modA_C" for module A and grismC
		# cross_filter: "F444W"
		# config_path: "/Users/GRISMDATA/NIRCAM/"
		config = os.path.join(config_path,"NIRCAM_%s_%s.conf" % (cross_filter,mode))
		#print config
		self.config = config
		
		self.image_seeds = image_seeds
		self.cross_filter = cross_filter
		self.mode = mode
		self.config_path = config_path

		# Get information about input padding. We use the first image seed for this, just like for the segmentation info.
		h = fits.open(image_seeds[0])[0].header
		self.xstart = h["NOMXSTRT"]
		self.xend = h["NOMXEND"]
		self.ystart = h["NOMYSTRT"]
		self.yend = h["NOMYEND"]

		#print(self.xstart,self.xend,self.xend-self.xstart)

		# Get segmentation info, from the first image seed.
		self.seg_data = fits.open(image_seeds[0])[2].data

		self.extrapolate_SED = extrapolate_SED
		self.SED_file = SED_file

	def observation(self,orders=["+1","+2"],max_split=100):
		# order: dispersion order, e.g. "+1"
		# max_split: we use max_split groups of pixels to disperse the whole image
		self.this_one = {}
		for order in orders:
			self.this_one[order] = NIRCAM_Gsim_observation(self.image_seeds,self.seg_data,self.config,order=order,max_split=max_split,extrapolate_SED=self.extrapolate_SED,SED_file=self.SED_file)
			self.this_one[order].disperse_all()

	def finalize(self,tofits=None,Back=False):
		if Back!=False:
			#final = fits.open(os.path.join(self.config_path,"%s_%s_back_V2.1.fits" % (self.cross_filter,self.mode)))[0].data
			final = fits.open(os.path.join(self.config_path,Back))[0].data
		else:
			final = 0.
		for order in self.this_one.keys():
			print(order)
			sim = self.this_one[order].simulated_image[self.ystart:self.yend+1,self.xstart:self.xend+1]
			final = final + sim
		self.final = final	
		if tofits!=None:
			self.saveSingleFits(tofits)

	def saveSingleFits(self,name):
		#save an array into the first extension of a fits file
		h0 = fits.PrimaryHDU()
		h1 = fits.ImageHDU(self.final,name='DATA')
        
        
		hdulist = fits.HDUList([h0,h1])
		hdulist.writeto(name,overwrite=True)

if __name__ == '__main__':
	import glob

	image_seeds = glob.glob("V4*.fits")
	seed = Grism_seed(image_seeds,"F444W","modA_R","/Users/npirzkal/Dropbox/GRISMDATA/NIRCAM/")






import os
import sys
from glob import glob
import numpy.ctypeslib as npct
import numpy as np
import ctypes
from ctypes import c_int

#print __file__
this_path = os.path.split(__file__)[0]
#print(this_path)

so_file = glob(os.path.join(this_path,'polyclip_c*.so'))
#print(so_file)
if len(so_file) >= 1:
    so_file = so_file[0]
else:
    print("WARNING: Cannot find polyclip_c*.so library")
    sys.exit()
polyclip = ctypes.cdll.LoadLibrary(so_file)
#polyclip = ctypes.cdll.LoadLibrary(os.path.join(this_path,"polyclip_c.so"))

array_1d_int_l = npct.ndpointer(dtype=np.int32, ndim=1, flags='CONTIGUOUS')
array_1d_int_r = npct.ndpointer(dtype=np.int32, ndim=1, flags='CONTIGUOUS')
array_1d_int_b = npct.ndpointer(dtype=np.int32, ndim=1, flags='CONTIGUOUS')
array_1d_int_t = npct.ndpointer(dtype=np.int32, ndim=1, flags='CONTIGUOUS')

array_1d_double_px = npct.ndpointer(dtype=np.float32, ndim=1, flags='CONTIGUOUS')
array_1d_double_py = npct.ndpointer(dtype=np.float32, ndim=1, flags='CONTIGUOUS')
array_1d_double_px_out = npct.ndpointer(dtype=np.float32, ndim=1, flags='CONTIGUOUS')
array_1d_double_py_out = npct.ndpointer(dtype=np.float32, ndim=1, flags='CONTIGUOUS')
array_1d_double_ri_out = npct.ndpointer(dtype=np.float32, ndim=1, flags='CONTIGUOUS')
array_1d_double_areas = npct.ndpointer(dtype=np.float32, ndim=1, flags='CONTIGUOUS')
array_1d_double_nclip_poly = npct.ndpointer(dtype=np.int32, ndim=1, flags='CONTIGUOUS')
array_1d_int_poly_inds = npct.ndpointer(dtype=np.int32, ndim=1, flags='CONTIGUOUS')
array_1d_double_inds = npct.ndpointer(dtype=np.int32, ndim=2, flags='CONTIGUOUS')
array_1d_double_x = npct.ndpointer(dtype=np.int32, ndim=1, flags='CONTIGUOUS')
array_1d_double_y = npct.ndpointer(dtype=np.int32, ndim=1, flags='CONTIGUOUS')
array_1d_double_index = npct.ndpointer(dtype=np.int32, ndim=1, flags='CONTIGUOUS')
polyclip.polyclip_multi2.restype = None

polyclip.polyclip_multi4.argtypes = [array_1d_int_l, # l
                                        array_1d_int_r, # r
                                        array_1d_int_b, # b
                                        array_1d_int_t, # t
                                        array_1d_double_px, # px
                                        array_1d_double_py, # py
                                        c_int, # n_poly
                                        array_1d_int_poly_inds, # poly_inds
                                        array_1d_double_x, # x
                                        array_1d_double_y, # y
                                        array_1d_double_nclip_poly, # nclip_poly
                                        array_1d_double_areas, # areas
                                        array_1d_double_index # output index
                                        ]




import numpy as np
import grismconf
import os
from astropy.io import fits
from scipy import sparse
from astropy.table import Table
#from NIRCAM_Gsim import polyclip
#from NIRCAM_Gsim.disperse import *
from ..disperse.disperse import dispersed_pixel
import h5py
from scipy.interpolate import interp1d
import platform
import multiprocessing
#import ray
from tqdm import tqdm
import psutil
import pickle

class interp1d_picklable(object):
    """ class wrapper for piecewise linear function
    """

    def __init__(self, xi, yi, **kwargs):
        self.xi = xi
        self.yi = yi
        self.args = kwargs
        self.f = interp1d(xi, yi, **kwargs)

    def __call__(self, xnew):
        return self.f(xnew)

    def __getstate__(self):
        return self.xi, self.yi, self.args

    def __setstate__(self, state):
        self.f = interp1d(state[0], state[1], **state[2])

def comprehension_flatten( aList ):
        return list(y for x in aList for y in x)

def helper(vars):
    x0s,y0s,f,order,C,ID,extrapolate_SED, xoffset, yoffset, outbound = vars # in this case ID is dummy number
    p = dispersed_pixel(x0s,y0s,f,order,C,ID,extrapolate_SED=extrapolate_SED,xoffset=xoffset,yoffset=yoffset,outbound=outbound)
    xs, ys, areas, lams, counts,ID = p
    IDs = [ID] * len(xs)

    pp = np.array([xs, ys, areas, lams, counts,IDs])
    return pp


# @ray.remote
# def mega_helper(pars):
#     return [helper(p) for p in pars]
    


class observation():
    # This class defines an actual observations. It is tied to a single flt and a single config file
    
    def __init__(self,mypool,direct_images,segmentation_data,config,mod="A",order="+1",plot=0,max_split=100,SED_file=None,extrapolate_SED=False,max_cpu=-1,ID=0,SBE_save=None, boundaries=[], outbound=False,renormalize=True, resample=True, multiprocessor='multiprocessing',seed_cube=None,save_objects=None):
        """direct_images: List of file name containing direct imaging data
        segmentation_data: an array of the size of the direct images, containing 0 and 1's, 0 being pixels to ignore
        config: The path and name of a GRISMCONF NIRCAM configuration file
        mod: Module, A or B
        order: The name of the spectral order to simulate, +1 or +2 for NIRCAM
        max_split: Number of chunks to compute instead of trying everything at once.
        SED_file: Name of HDF5 file containing datasets matching the ID in the segmentation file and each consisting of a [[lambda],[flux]] array.
        SBE_save: If set to a path, HDF5 containing simulated stamps for all obsjects will be saved.
        boundaries: a tuple containing the coordinates of the FOV within the larger seed image.
        multiprocessor: 'multiprocessing' for native Python.
        """
        print("direct_images:",direct_images)
        if not seed_cube is None:
            print("Seed images: 3D cube")
        #print("segmentation_data:",segmentation_data)
        print("config:",config)
        print("mod:",mod)
        print("order:",order)
        print("plot:",plot)
        print("max_split:",max_split)
        print("SED_file:",SED_file)
        print("extrapolate_SED:",extrapolate_SED)
        print("max_cpu:",max_cpu)
        print("ID:",ID)
        print("SBE_save:",SBE_save)
        print("boundaries:",boundaries)
        print("renormalize:",renormalize)
        print("resample:",resample)
        print("multiprocessor:",multiprocessor)


        if max_cpu<0:
            max_cpu = psutil.cpu_count(logical=False) 

        print("Loading grism configuration file:",config)
        self.C = grismconf.Config(config)
            
        if plot:
            import matplotlib.pyplot as plt
            plt.ion()
            plt.clf()
            x = np.arange(self.C.WMIN,self.C.WMAX,(self.C.WMAX,self.C.WMIN)/100.)
            plt.plot(x,self.C.SENS[order](x))

        self.ID = ID
        self.IDs = []
        self.dir_images = direct_images
        self.seg = segmentation_data
        self.c3d = seed_cube
        self.dims = np.shape(self.seg)
        self.outbound = outbound
        self.order = order
        self.SED_file = SED_file
        self.SBE_save = SBE_save
        self.max_cpu = max_cpu
        self.cache = False
        self.POM_mask = None
        self.POM_mask01 = None
        self.renormalize = renormalize
        self.resample = resample
        self.multiprocessor = multiprocessor 
        self.mypool = mypool
        self.save_objects = save_objects

        print("Ooutbound is",outbound)

        if self.C.POM is not None:
            print("Using POM mask",self.C.POM)
            with fits.open(self.C.POM) as fin:
                self.POM_mask = fin[1].data
                self.POM_mask01 = fin[1].data * 1. # A version of the mask that only contains 0-1 values. Pixels with labels/values>10000 are set to 1.
                self.POM_mask01[self.POM_mask01>=10000] = 1.
                self.Pxstart = int(fin[1].header["NOMXSTRT"])
                self.Pxend = int(fin[1].header["NOMXEND"])
                self.Pystart = int(fin[1].header["NOMYSTRT"])
                self.Pyend = int(fin[1].header["NOMYEND"])

                if len(fin)>2:
                    self.POM_transmission = {}
                    for i in range(2,len(fin)):
                        try:
                            w = fin[i].data["Wavelength"]
                            t = fin[i].data["Transmission"]
                            indx = int(fin[i].header["POMINDX"])
                            self.POM_transmission[indx] = interp1d_picklable(w,t,bounds_error=False,fill_value=0.)
                            print("Loading extra optical element transmission curves from POM file extension {}".format(i))
                        except:
                            pass

        if len(boundaries)!=4:           
            xpad = (np.shape(segmentation_data)[1]-self.C.NAXIS[0])//2
            ypad = (np.shape(segmentation_data)[0]-self.C.NAXIS[1])//2
            self.xstart = 0 + xpad
            self.xend = xpad + self.C.NAXIS[0]-1
            self.ystart = 0 + ypad
            self.yend = ypad + self.C.NAXIS[1]-1
            print("No boundaries passed. Assuming symmetrical padding of {} {} pixels and a final size of {} {} .".format(xpad,ypad,self.xend+1-self.xstart,self.yend+1-self.ystart))
        else:
            self.xstart,self.xend,self.ystart,self.yend = boundaries

        
        self.extrapolate_SED = extrapolate_SED # Allow for SED extrapolation
        if self.extrapolate_SED:
            print("Warning: SED Extrapolation turned on.")
        if self.renormalize is False:
            print("Warning: not re-normalizing sources to unity")
            
        # if self.multiprocessor=='ray':
        #     print("Using ray for multiprocessing")
        #     ray.init(num_cpus=self.max_cpu,ignore_reinit_error=True)

        self.apply_POM()
        # self.create_pixel_list()
        
        self.p_l = []
        self.p_a = []


    def apply_POM(self):
        """Account for the finite size of the POM and remove pixels in segmentation files which should not
        be dispersed.
        If a POM mask array was loaded then we make sure it is modified to have the same size as the input seg map
        and with the detector FOV starting at the same pixel locations as the input seg map."""

        if self.POM_mask is None:
            x0 = int(self.xstart+self.C.XRANGE[self.C.orders[0]][0] + 0.5)
            x1 = int(self.xend+self.C.XRANGE[self.C.orders[0]][1] + 0.5)
            y0 = int(self.ystart+self.C.YRANGE[self.C.orders[0]][0] + 0.5)
            y1 = int(self.yend+self.C.YRANGE[self.C.orders[0]][1] + 0.5)

            if x0<0: x0 = 0
            if y0<0: y0 = 0
            ymax,xmax = np.shape(self.seg)
            if x0+1>xmax: x0 = xmax-1
            if y0+1>ymax: y0 = ymax-1

            from astropy.io import fits
            print("POM footprint applied: [{}:{},{}:{}]".format(x0,x1,y0,y1))
            print("Pixels outside of this region of the input data ([{},{}]) will not be dispersed.".format(xmax,ymax))
            self.seg[:,:x0+1] = 0
            self.seg[:,x1:] = 0
            self.seg[:y0+1,:] = 0
            self.seg[y1:,:] = 0
            
            if not self.c3d is None:
                for k in self.c3d.keys():
                    sx0,sy0,dsata,sseg = self.c3d[k]
                    sx1 = sx0 + np.shape(sseg)[1]
                    sy1 = sy0 + np.shape(sseg)[0]
                    self.c3d[k][3] = self.c3d[k][3] * self.seg[sy0:sy1,sx0:sx1]

        else:
            self.Pxstart - self.xstart
            self.Pystart - self.ystart

            ys1,xs1 = np.shape(self.POM_mask)
            ys2,xs2 = np.shape(self.seg)

            #print("xstart ettc:",self.xstart,self.xend,self.ystart,self.yend)
            #print("POM start etc:",self.Pxstart ,self.Pxend,self.Pystart ,self.Pyend)
            #print(self.Pystart - self.ystart,self.yend-self.ystart,self.Pxstart - self.xstart+1,self.xend-self.xstart+1)
            if (xs1>xs2) or (ys1>ys2):
                print("Warning: The input seg map is smaller ({},{}) than the POM mask ({},{}) for this mode.".format(xs2,ys2,xs1,ys1))
                raise Exception("Invalid seg map size.")

            # Crate a new POM mask that is the size of the input seg and data arrays
            # Compute the offset between where the detector is in the original POM mask and the input sef file
            xoff = self.xstart - self.Pxstart
            yoff = self.ystart - self.Pystart

            if (xoff<0) or (yoff<0):
                print("Warning: the input seg map and POM mask are not compatible. The seg map needs a larger border around the detector FOV than the POM mask.")
                print("seg map detector FOV starts at {} {}".format(self.xstart,self.ystart))
                print("POM mask detector FOV starts at {} {}".format(self.Pxstart,self.Pystart))
                raise Exception("Invalid seg map size.")
 
            POM = np.zeros(np.shape(self.seg))
            POM[yoff:yoff+ys1,xoff:xoff+xs1] = self.POM_mask
            self.POM_mask = POM*1

            # POM = np.zeros(np.shape(self.seg))
            # POM[yoff:yoff+ys1,xoff:xoff+xs1] = self.POM_mask
            POM[POM>1] = 1
            self.POM_mask01 = POM*1

            self.Pxstart  = self.xstart
            self.Pxend  = self.xend
            self.Pystart  = self.ystart
            self.Pyend  = self.ystart

            # We apply the POM mask to the seg file, but only as a mask, romving pixels getting no lights. We keep the POM mask with its orginal values otherwise 
            # to be able to account for partial transmission later on.
    

            #from astropy.io import fits
            #print("POM size:",np.shape(POM))
            #print("seg size:",np.shape(self.seg))
            #fits.writeto("seg_org.fits",self.seg,overwrite=True)
            mask = np.zeros(np.shape(self.POM_mask01),int)
            mask[self.POM_mask01>0] = 1
            self.seg = self.seg * mask
            #fits.writeto("POM_msk.fits",self.POM_mask,overwrite=True)
            #fits.writeto("POM.fits",self.POM_mask01 ,overwrite=True)
            #fits.writeto("seg_msk.fits",self.seg,overwrite=True)
            if not self.c3d is None:
                for k in self.c3d.keys():
                    sx0,sy0,dsata,sseg = self.c3d[k]
                    sx1 = sx0 + np.shape(sseg)[1]
                    sy1 = sy0 + np.shape(sseg)[0]
                    self.c3d[k][3] = self.c3d[k][3] * mask[sy0:sy1,sx0:sx1]

    def create_pixel_list(self):
        # Create a list of pixels to dispersed, grouped per object ID



        if not self.c3d is None:
            print("Using seed cube")
            self.create_pixel_list_from_seed_cube()
            return

        if self.save_objects is not None and os.path.isfile(self.save_objects):
            print("Loading from pre-computed ",self.save_objects)
            self.IDs,self.xs,self.ys,self.fs = pickle.load(open(self.save_objects,"rb"))
            return

        if self.ID==0:
            self.xs = []
            self.ys = []
            all_IDs = np.array(list(set(np.ravel(self.seg))))
            all_IDs = all_IDs[all_IDs>0]

            # Remove
            # self.IDs = self.IDs[0:10]
            # all_IDs = all_IDs[0:10]
            #

            print("We have ",len(all_IDs),"Objects")
            for ID in tqdm(all_IDs):
                ys,xs = np.nonzero(self.seg==ID)

                if (len(xs)>0) & (len(ys)>0):
                    self.xs.append(xs)
                    self.ys.append(ys)
                    self.IDs = all_IDs
        else:
            vg = self.seg==self.ID
            ys,xs = np.nonzero(vg)            
           
            if (len(xs)>0) & (len(ys)>0):    
                self.xs.append(xs)
                self.ys.append(ys)
                self.IDs = [self.ID]

        self.fs = {}
        for dir_image in self.dir_images:
            print("dir image:",dir_image)
            if self.SED_file==None:
                try:
                    l = fits.getval(dir_image,'PHOTPLAM') / 10000. # in Angsrrom and we want Micron now
                except:
                    print("WARNING: unable to find PHOTPLAM keyword in {}".format(dir_image))
                    sys.exit()

                try:
                    photflam = fits.getval(dir_image,'photflam')
                except:
                    print("WARNING: unable to find PHOTFLAM keyword in {}".format(dir_image))
                    sys.exit()
                print("Loaded",dir_image, "wavelength:",l,"micron")

            try:
                d = fits.open(dir_image)[1].data
            except:
                d = fits.open(dir_image)[0].data

            # If we do not use an SED file then we use photometry to get fluxes
            # Otherwise, we assume that objects are normalized to 1.
            if self.SED_file==None:
                self.fs[l] = []
                dnew = d
                if self.POM_mask01 is not None:
                    dnew = d * self.POM_mask01 # Apply POM transmission mask to the data pixels

                for i in range(len(self.IDs)):
                    self.fs[l].append(dnew[self.ys[i],self.xs[i]] * photflam)

            else:
                # Need to normalize the object stamps              
                for ID in tqdm(self.IDs,desc='Normalizing objects footprint'):
                    vg = self.seg==ID
                    dnew = d
                    
                    if self.renormalize is True:
                        sum_seg = np.sum(dnew[vg]) # But normalize by the whole flux
                        if sum_seg!=0.:
                            dnew[vg] = dnew[vg]/sum_seg
                    # else:
                    #     print("not re-normalizing sources to unity")

                    if self.POM_mask01 is not None:
                        #print("Applying POM transmission to data")
                        dnew = dnew * self.POM_mask01 # Apply POM transmission mask to the data pixels. This is a single grey correction for the whole object.

                self.fs["SED"] = []
                for i in range(len(self.IDs)):
                    self.fs["SED"].append(dnew[self.ys[i],self.xs[i]])
        if self.save_objects is not None:
            print("Saving pre-computed objects to ",self.save_objects)
            pickle.dump([self.IDs, self.xs,self.ys,self.fs],open(self.save_objects,"wb"))

        print("Done")

    def create_pixel_list_from_seed_cube(self):
        # Create a list of pixels to dispersed, grouped per object ID
        dir_image = self.dir_images[0]
        if self.SED_file==None:
            try:
                l = fits.getval(dir_image,'PHOTPLAM') / 10000. # in Angsrrom and we want Micron now
            except:
                print("WARNING: unable to find PHOTPLAM keyword in {}".format(dir_image))
                sys.exit()

            try:
                photflam = fits.getval(dir_image,'photflam')
            except:
                print("WARNING: unable to find PHOTFLAM keyword in {}".format(dir_image))
                sys.exit()
            print("Loaded",dir_image, "wavelength:",l,"micron")
            
        if self.ID==0:
            self.IDs = np.array(list(set(np.ravel(self.seg))))
            self.IDs = self.IDs[self.IDs>0]
        else:
            self.IDs = [self.ID]


        self.xs = []
        self.ys = []
        self.fs = {}

        if self.SED_file==None:
            self.fs[l] = []
        else:
            self.fs["SED"] = []

        print("We have ",len(self.IDs),"Objects")
        for ID in self.IDs:
            data_stp = self.c3d[ID][2]*1
            seg_stp = self.c3d[ID][3]*1
            ys0,xs0 = np.nonzero(seg_stp>0)
            xs = xs0 + self.c3d[ID][0]
            ys = ys0 + self.c3d[ID][1]
            # Coords in full frame
            xmin = self.c3d[ID][0]
            ymin = self.c3d[ID][1]
            xmax = xmin + np.shape(data_stp)[1]
            ymax = ymin + np.shape(data_stp)[0]

            self.xs.append(xs)
            self.ys.append(ys)
        
            if self.POM_mask01 is not None:
                data_stp = data_stp * self.POM_mask01[ymin:ymax,xmin:xmax]

            # If we do not use an SED file then we use photometry to get fluxes
            # Otherwise, we assume that objects are normalized to 1.
            if self.SED_file==None:
                    self.fs[l].append(data_stp[ys0,xs0] * photflam)
            else:
                # Need to normalize the object stamps                                  
                if self.renormalize is True:
                    sum_data_stp = np.sum(data_stp) # But normalize by the whole flux
                    if sum_data_stp!=0.:
                        data_stp = data_stp/sum_data_stp
                self.fs["SED"].append(data_stp[ys0,xs0])
    
    def disperse_all(self,cache=False):

        if cache:
            print("Object caching ON")
            self.cache = True
            self.cached_object = {}

        self.simulated_image = np.zeros(self.dims,float)
        
        self.create_pixel_list()
        
        for i in tqdm(range(len(self.IDs)),desc='Dispersing order {}'.format(self.order)):
            if self.cache:
                self.cached_object[i] = {}
                self.cached_object[i]['x'] = []
                self.cached_object[i]['y'] = []
                self.cached_object[i]['f'] = []
                self.cached_object[i]['w'] = []
                self.cached_object[i]['minx'] = []
                self.cached_object[i]['maxx'] = []
                self.cached_object[i]['miny'] = []
                self.cached_object[i]['maxy'] = []

            this_object = self.disperse_chunk(i)

            if self.SBE_save != None:
                # If SBE_save is enabled, we create an HDF5 file containing the stamp of this simulated object
                # order is in self.order
                # We just save the x,y,f,w arrays as well as info about minx,maxx,miny,maxy
    
                # We trim the stamp to avoid padding area
                this_SBE_object =  this_object[self.ystart:self.yend+1,self.xstart:self.xend+1]
                
                yss,xss = np.nonzero(this_SBE_object>0)
                
                if len(xss)<1:
                    continue 

                minx = np.min(xss)
                maxx = np.max(xss)
                miny = np.min(yss)
                maxy = np.max(yss)

                this_SBE_object = this_SBE_object[miny:maxy+1,minx:maxx+1]

                if os.path.isfile(self.SBE_save):
                    mode = "a"
                else:
                    mode = "w"

                with h5py.File(self.SBE_save,mode) as fhdf5:
                    dset = fhdf5.create_dataset("%d_%s" % (self.IDs[i],self.order),data=this_SBE_object,dtype='f',compression="gzip",compression_opts=9)
                    dset.attrs[u'minx'] = minx
                    dset.attrs[u'maxx'] = maxx
                    dset.attrs[u'miny'] = miny
                    dset.attrs[u'maxy'] = maxy
                    dset.attrs[u'units'] = 'e-/s'


    def disperse_background_1D(self,background):
        """Method to create a simple disperse background, obtained by dispersing a full row or column.
        We assume no field dependence in the cross dispersion direction and create a full 2D image by tiling a single dispersed row or column"""

        # Create a fake object, line in middle of detector
        C = self.C
        naxis = [self.xend-self.xstart+1 ,self.yend-self.ystart+1] 
        xpos,ypos = naxis[0]//2,naxis[1]//2

        # Find out if this an x-direction or y-direction dispersion
        dydx = np.array(C.DISPXY(self.order,1000,1000,1))-np.array(C.DISPXY(self.order,1000,1000,0))
        if np.abs(dydx[0])>np.abs(dydx[1]):
            #print("disperse_background_1D: x-direction")
            direction = "x"
            xs = np.arange(self.C.XRANGE[self.order][0]+0,self.C.XRANGE[self.order][1]+naxis[0])
            ys = np.zeros(np.shape(xs))+ypos
        else:
            #print("disperse_background_1D: y-direction")
            direction = "y"
            ys = np.arange(self.C.YRANGE[self.order][0]+0,self.C.YRANGE[self.order][1]+naxis[0])
            xs = np.zeros(np.shape(ys))+xpos

        #print(xpos,ypos)
        
        
        lam = background[0]
        fnu = background[1]

        fnu = fnu/4.25e10 # MJy/arcsec^2
        fnu = fnu*1e6 # Jy/arcsec^2
        fnu = fnu * (0.065**2) # Jy/pixel

        fnu = fnu*1e-23
        c = 299792458.* 1e10 # A
        wa = lam*10000
        flam = fnu/(wa**2/c)

        f = [lam,flam]
                
        pars = []        
        for i in range(len(xs)):
            ID = 1
            xs0 = [xs[i],xs[i]+1,xs[i]+1,xs[i]]
            ys0 = [ys[i],ys[i],ys[i]+1,ys[i]+1]
            pars.append([xs0,ys0,f,self.order,C,ID,False,self.xstart,self.ystart,self.outbound])


        bck = np.zeros(naxis,float)

        chunksize = int(len(pars)/self.max_cpu)
        if chunksize<1:
            chunksize = 1
        # if self.multiprocessor=='ray':
        #     #ray.init(num_cpus=self.max_cpu,ignore_reinit_error=True)
            
        #     chunked_pars = [pars[i * chunksize:(i + 1) * chunksize] for i in range((len(pars) + chunksize - 1) // chunksize )] 

        #     result_ids = []
        #     [result_ids.append(mega_helper.remote(cpars)) for cpars in chunked_pars]

        #     results = ray.get(result_ids)
        #     results = [item for sublist in results for item in sublist]

        #     del result_ids
        #     ##ray.shutdown()
        # else:
        for pp in self.mypool.imap_unordered(helper, pars, chunksize=chunksize):
            if np.shape(pp.transpose())==(1,6):
                continue
            
            x,y,w,f = pp[0],pp[1],pp[3],pp[4]

            vg = (x>=0) & (x<naxis[0]) & (y>=0) & (y<naxis[1]) 

            x = x[vg]
            y = y[vg]
            f = f[vg]
            w = w[vg]
                    
            if len(x)<1:
                continue

            minx = int(min(x))
            maxx = int(max(x))
            miny = int(min(y))
            maxy = int(max(y))
            a = sparse.coo_matrix((f, (y-miny, x-minx)), shape=(maxy-miny+1,maxx-minx+1)).toarray()
            bck[miny:maxy+1,minx:maxx+1] = bck[miny:maxy+1,minx:maxx+1] + a

        if direction=="x":
            bck = np.sum(bck,axis=0)
            bck = np.tile(bck,[naxis[1],1])
        else:
            bck = np.sum(bck,axis=1)
            bck = np.tile(bck,[naxis[0],1]).transpose()

        return bck

    def smooth(self,w,f,dw=0.002):
        from scipy.interpolate import interp1d
        from astropy.convolution import Gaussian1DKernel
        from astropy.convolution import convolve
        fct = interp1d(w,f)
        xs = np.arange(np.min(w),np.max(w),1./10000)
        ys = fct(xs)
        kernel = Gaussian1DKernel(stddev=dw)
        conv = convolve(ys,kernel)
        fct2 = interp1d(xs,conv)
        xs2 = np.arange(np.min(xs),np.max(xs),dw)
        ys2 = fct2(xs2)
        return xs2,ys2

    def disperse_chunk(self,c):
        """Method that handles the dispersion. To be called after create_pixel_list()"""
        import time


        if self.SED_file!=None:
            # We use an input spectrum file
            import h5py
            with h5py.File(self.SED_file,'r') as h5f:
                pars = []
                ###ID = int(self.seg[self.ys[c][0],self.xs[c][0]])
                ID = int(self.IDs[c])
                tmp = h5f["%s" % (ID)][:]

                if self.resample:
                    

                    # Figuring out a few things about size of order, dispersion and wavelengths to use    
                    wmin = self.C.WRANGE[self.order][0]
                    wmax = self.C.WRANGE[self.order][1]

                    t0 = self.C.INVDISPL(self.order,self.C.NAXIS[0]/2,self.C.NAXIS[1]/2,wmin)
                    t1 = self.C.INVDISPL(self.order,self.C.NAXIS[0]/2,self.C.NAXIS[1]/2,wmax)
                    
                    dx0 = self.C.DISPX(self.order,self.C.NAXIS[0]/2,self.C.NAXIS[1]/2,t0) - self.C.DISPX(self.order,self.C.NAXIS[0]/2,self.C.NAXIS[1]/2,t1)
                    dx1 = self.C.DISPY(self.order,self.C.NAXIS[0]/2,self.C.NAXIS[1]/2,t0) - self.C.DISPY(self.order,self.C.NAXIS[0]/2,self.C.NAXIS[1]/2,t1)

                    dw = np.abs((wmax-wmin)/(dx1-dx0))
                    print("Smoothing and resampling input spectrum to {} micron".format(dw))
                    # print("self.SED_file:",self.SED_file)
                    # print("lams:",tmp[0])
                    # print(tmp[0][1:]-tmp[0][:-1])
                    # print("DW:",dw)

                    tmp = self.smooth(tmp[0],tmp[1],dw/1.5)
                    #print(len(tmp[0]),len(tmp[1]))
                    
                
                ok = (tmp[0]>self.C.WRANGE[self.order][0]) & (tmp[0]<self.C.WRANGE[self.order][1])                        
                tmp = [tmp[0][ok],tmp[1][ok]]

                for i in range(len(self.xs[c])):
                    
                    lams = tmp[0]
                    fffs = tmp[1]*self.fs["SED"][c][i]
   
                    # trim input spectrum 
                    # try:
                    #     ok = (lams>self.C.WRANGE["+1"][0]) & (lams<self.C.WRANGE["+1"][1])
                    # except:
                    #     ok = (lams>self.C.WRANGE["A"][0]) & (lams<self.C.WRANGE["A"][1])

                    # ok = (lams>self.C.WRANGE[self.order][0]) & (lams<self.C.WRANGE[self.order][1])                        
                    # lams = lams[ok]
                    # fffs = fffs[ok]
                    #print(len(lams))
                    #sys.exit(1)
                    if self.POM_mask is not None:
                        POM_value = self.POM_mask[self.ys[c][i],self.xs[c][i]]
                    else:
                        POM_value = 1.
                    if POM_value>=10000:
                        #print("Applying additional transmission of :",self.xs[c][i],self.ys[c][i],POM_value)
                        trans = self.POM_transmission[POM_value](lams)
                        fffs = fffs * trans
                    else:
                        fffs = fffs * POM_value

                    if len(lams)>0:
                        f = [lams,fffs]
                        xs0 = [self.xs[c][i],self.xs[c][i]+1,self.xs[c][i]+1,self.xs[c][i]]
                        ys0 = [self.ys[c][i],self.ys[c][i],self.ys[c][i]+1,self.ys[c][i]+1]
                        pars.append([xs0,ys0,f,self.order,self.C,ID,self.extrapolate_SED,self.xstart,self.ystart,self.outbound])
        else:
            # No spectrum passed
            pars = []
            for i in range(len(self.xs[c])):
                ID = i
                xs0 = [self.xs[c][i],self.xs[c][i]+1,self.xs[c][i]+1,self.xs[c][i]]
                ys0 = [self.ys[c][i],self.ys[c][i],self.ys[c][i]+1,self.ys[c][i]+1]
                lams = np.array(list(self.fs.keys()))
                flxs = np.array([self.fs[l][c][i] for l in self.fs.keys()])
                ok = flxs!=0 # We avoid any pixel containing pure 0's
                if len(flxs[ok])==0: continue
                flxs = flxs[ok]
                lams = lams[ok]
                ok = np.argsort(lams)
                flxs = flxs[ok]
                lams = lams[ok]

                if self.POM_mask is not None:
                    POM_value = self.POM_mask[self.ys[c][i],self.xs[c][i]]
                else:
                    POM_value = 1.
                if POM_value>10000:
                    #print("Applying additional transmission of :",self.xs[c][i],self.ys[c][i],POM_value)
                    trans = self.POM_transmission[POM_value](lams)
                    flxs = flxs * trans
                else:
                    flxs = flxs * POM_value

                if len(lams)>0:
                    f = [lams,flxs]
                    pars.append([xs0,ys0,f,self.order,self.C,ID,self.extrapolate_SED,self.xstart,self.ystart,self.outbound])

        time1 = time.time()

        this_object = np.zeros(self.dims,float)
        chunksize = int(len(pars)/self.max_cpu)
        if chunksize<1:
            chunksize = 1

        chunksize = 10
        #print(len(pars),self.max_cpu,chunksize)
        if self.multiprocessor=='ray':
#            ray.init(num_cpus=self.max_cpu,ignore_reinit_error=True)

            chunked_pars = [pars[i * chunksize:(i + 1) * chunksize] for i in range((len(pars) + chunksize - 1) // chunksize )] 

            result_ids = []
            [result_ids.append(mega_helper.remote(cpars)) for cpars in chunked_pars]
            
            results = ray.get(result_ids)
            results = [item for sublist in results for item in sublist]

            del result_ids
            #ray.shutdown()
        else:
            #with multiprocessing.Pool(processes=self.max_cpu) as mypool:
            for pp in self.mypool.imap_unordered(helper, pars, chunksize=chunksize):
                if np.shape(pp.transpose())==(1,6):
                    continue
    
                x,y,w,f = pp[0],pp[1],pp[3],pp[4]

                vg = (x>=0) & (x<self.dims[1]) & (y>=0) & (y<self.dims[0]) & (np.isfinite(f)) 

                x = x[vg]
                y = y[vg]
                f = f[vg]
                w = w[vg]
                    
                if len(x)<1:
                    continue
                    
                minx = int(min(x))
                maxx = int(max(x))
                miny = int(min(y))
                maxy = int(max(y))
                a = sparse.coo_matrix((f, (y-miny, x-minx)), shape=(maxy-miny+1,maxx-minx+1)).toarray()
                self.simulated_image[miny:maxy+1,minx:maxx+1] = self.simulated_image[miny:maxy+1,minx:maxx+1] + a
                this_object[miny:maxy+1,minx:maxx+1] = this_object[miny:maxy+1,minx:maxx+1] + a

                if self.cache:
                    self.cached_object[c]['x'].append(x)
                    self.cached_object[c]['y'] .append(y)
                    self.cached_object[c]['f'].append(f)
                    self.cached_object[c]['w'].append(w)
                    self.cached_object[c]['minx'].append(minx)
                    self.cached_object[c]['maxx'].append(maxx)
                    self.cached_object[c]['miny'].append(miny)
                    self.cached_object[c]['maxy'].append(maxy)
            
        time2 = time.time()

        return this_object

    def disperse_all_from_cache(self,trans=None):
        if not self.cache:
            print("No cached object stored.")
            return

        self.simulated_image = np.zeros(self.dims,float)

        for i in tqdm(range(len(self.IDs)),desc="Dispersing order {} from cache".format(self.order)):
            this_object = self.disperse_chunk_from_cache(i,trans=trans)


    def disperse_chunk_from_cache(self,c,trans=None):
        """Method that handles the dispersion. To be called after create_pixel_list()"""
        
        import time

        if not self.cache:
            print("No cached object stored.")
            return

        time1 = time.time()
        
        this_object = np.zeros(self.dims,float)

        if trans!=None:
                print("Applying a transmission function...")
        for i in range(len(self.cached_object[c]['x'])): 
            x = self.cached_object[c]['x'][i]
            y = self.cached_object[c]['y'][i]
            f = self.cached_object[c]['f'][i]*1.
            w = self.cached_object[c]['w'][i]

            if trans!=None:
                f *= trans(w)

            minx = self.cached_object[c]['minx'][i]
            maxx = self.cached_object[c]['maxx'][i]
            miny = self.cached_object[c]['miny'][i]
            maxy = self.cached_object[c]['maxy'][i]
   
            a = sparse.coo_matrix((f, (y-miny, x-minx)), shape=(maxy-miny+1,maxx-minx+1)).toarray()
            self.simulated_image[miny:maxy+1,minx:maxx+1] = self.simulated_image[miny:maxy+1,minx:maxx+1] + a
            this_object[miny:maxy+1,minx:maxx+1] = this_object[miny:maxy+1,minx:maxx+1] + a

        time2 = time.time()

        print(time2-time1,"s.")
        return this_object

    def show(self):
        import matplotlib.pyplot as plt
        plt.ion()

        xx = self.p_x - min(self.p_x)
        yy = self.p_y - min(self.p_y)
        a = sparse.coo_matrix((self.p_f, (yy, xx)), shape=(max(yy)+1, max(xx)+1)).toarray()

        im = plt.imshow(a)
        im.set_clim(0,1)

        plt.draw()
        raw_input("...")




#! /usr/bin env python

"""
Code to generate WFSS dispersed seed images. Starting with a set of imaging seed images, 
potentially with padding, given a JWST WFSS GRISMCONF configuration file
"""

import os, pickle
from astropy.io import fits
import numpy as np
from .observations.observations \
import observation as Gsim_observation 
import multiprocessing

class Grism_seed():
    def __init__(self,image_seeds,cross_filter,mode,config_path=".",extrapolate_SED=False,SED_file=None,instrument="NIRCAM",max_cpu=None, SBE_save=None, renormalize=True, resample=False, outbound=False, save_objects=None):
        """A class for a grism simulation

        Attributes
        ----------
        image_seeds: list
            A list of image image_seeds or of seed cube files
        cross_filter: str
            A string containing the name of a direct filter
        mode: str
            A string containing either R or C
        config_path: str
            As string pointing to a GRISMCONF configuration file
        extrapolate_SED: bol
            If set to True, the objects' SED will be extrapolated if needed
        SED_file: str
            A string containing the name of an HDF5 file containing the spectra of each source
        instrument: str
            A string containing the name of the instrument (e.g. NIRCAM)
        max_cpu: int
            An integer containing the number of CPU to use in the multiheaded pool when dispersing
        SBE_save: str
            A string containing the name of an output HDF5 file which will contain simulated 2D stamps of each source
        renormalize: bol
            Whether to renormalize the input data to unity over segmentation map area when using an input spectrum.
        reaample: bol
            If true, the disperser will first smooth and resample input spectra appropriately to match the disperser resolution.
        save_objest: None or filename
            Save computed objects footprints for future use

        Methods
        -------
        observation(self,orders=["+1","+2"],max_split=-1000,ID=0)

        finalize(self,tofits=None,Back=False,BackLevel=None)

        saveSingleFits(self,name)

        """

        config = os.path.join(config_path,"%s_%s_%s.conf" % (instrument,cross_filter,mode))
        self.config = config
        
        self.image_seeds = image_seeds
        self.c3d = None
        self.cross_filter = cross_filter
        self.mode = mode
        self.config_path = config_path
        if max_cpu is None:
            max_cpu = multiprocessing.cpu_count() - 1

        self.max_cpu = max_cpu
        self.SBE_save = SBE_save

        if self.SBE_save != None:
            print("Will output to ", self.SBE_save)
            if os.path.isfile(self.SBE_save):
                os.unlink(self.SBE_save)


        # Get information about input padding. We use the first image seed for this, just like for the segmentation info.
        h = fits.open(image_seeds[0])[0].header
        self.xstart = int(h["NOMXSTRT"])
        self.xend = int(h["NOMXEND"])
        self.ystart = int(h["NOMYSTRT"])
        self.yend = int(h["NOMYEND"])

        # Get segmentation info, from the first image seed.
        self.seg_data = fits.open(image_seeds[0])[2].data

        self.extrapolate_SED = extrapolate_SED
        self.SED_file = SED_file
        self.renormalize = renormalize
        self.resample = resample
        self.outbound = outbound
        self.save_objects = save_objects

        print("Initializing Pool with",self.max_cpu,"cores")
        self.mypool =  multiprocessing.Pool(processes=self.max_cpu)

        # temp 
        # Check if we have seed cube and use these instead... we just load them 
        print("image_seeds:",image_seeds)

        if "_ptsrc_seed_image.fits" in image_seeds[0]:
            sc3d = image_seeds[0].split("_uncal")[0]+"_uncal_star_seed_cube.pickle"
            if os.path.isfile(sc3d):
                print("ptsrc seg seed data found")
                self.c3d = pickle.load(open(sc3d,"rb"))
        if "_extended_seed_image.fits" in image_seeds[0]:
            sc3d = image_seeds[0].split("_uncal.fits")[0]+"_uncal_galaxy_seed_cube.pickle"
            if os.path.isfile(sc3d):
                print("extended seg seed data found")
                self.c3d = pickle.load(open(sc3d,"rb"))

    def observation(self,orders=None,max_split=-1000,ID=0):
        """Sets up an observation.

        Parameters
        ----------
        orders : list
            A list of string containing the name of the orders to disperse
        max_split : int
            Maximum number of pixels to disperse at once (not currently used)
        ID : int
            Specific object ID to dispersed. Set to 0 (default) to disperse all of available objects
        """

        self.this_one = {}

        # If orders are not passed, we get them from the config file
        
        if orders==None:
            import grismconf
            C = grismconf.Config(self.config)
            print("orders:",C.orders)
            self.orders = C.orders
        else:
            self.orders = orders

        for order in self.orders:
            boundaries = [self.xstart,self.xend,self.ystart,self.yend]
            self.this_one[order] = Gsim_observation(self.mypool,self.image_seeds,self.seg_data,self.config,seed_cube=self.c3d,order=order,max_split=max_split,extrapolate_SED=self.extrapolate_SED,SED_file=self.SED_file,max_cpu=self.max_cpu,ID=ID, SBE_save=self.SBE_save,boundaries=boundaries,renormalize=self.renormalize,resample=self.resample, outbound=self.outbound, save_objects=self.save_objects)
            #self.this_one[order].disperse_all()

    def disperse(self,orders=None,cache=False,trans=None):
        """Run the disperser. 

        Parameters
        ----------
        orders: list
            Optional list containing the name of the orders to disperse
        cache: bool
            If set to True, the dispersion tables are cached and will be used on subsequent calls. 
        """
        if orders==None:
            orders = self.orders

        print("Dispersing orders ", orders)
        for order in orders:
            #print("Dispersing order ",order)
            if self.this_one[order].cache:
                self.this_one[order].disperse_all_from_cache(trans=trans)
            else:
                self.this_one[order].disperse_all(cache=cache)

    def disperse_background_1D(self,background):
        """Produces a dispersed 2D image of the background spectrum contained in the fits file background, meant to be the 
        output of thr jwst_background module. This background is dispersed either in the row or column direction, depending on the 
        dispersion, and the result is tiled to produce a full 2D image. All orders are generated and added up.

        Parameters
        ----------
        background: 2D numpy array [lambda values,flux values]
            A 2D numpy array containing the spectrum of the background to disperse. 
            The wavelength should be in (micron) and the flux (in Mjy/sr), as is produced by 
            the jwst_background package
        output: numpy 2D array
            A 2D array containing the model background which can be fed back into finalize()
        """

        bck = 0.
        for order in self.orders:
            print("Computing dispersed background for order ",order)
            bck += self.this_one[order].disperse_background_1D(background) 

#        fits.writeto("WFSS_background.fits",bck,overwrite=True)
        return bck

    def finalize(self,Back=None,BackLevel=None,tofits=None):
        """ Produces a 2D dispersed image and add the appropriate background

        Parameters
        ----------
        tofits: str 
            Name of a fits file to write the simulation to. Default is set to None
        Back: str
            Name of a fits file containing an image of the background in e-/s in extension 0
            If None, the file listed in the config file is used.
        BackLevel: float
            Renormalization factor for the background image. Renormalization is done by multiplying by BackLevel/np.median(Back)
        """

        # Initialize final image with the background estimate
        final = 0.
        if (Back is None) and (BackLevel is not None):
            # Use pre-computed background from config file, scaled by BackLevel
            import grismconf
            bck_file = grismconf.Config(self.config).BCK
            print("Adding pre-computed 2D dispersed background ",bck_file,"scaled to",BackLevel)
            import grismconf
            final = fits.open(bck_file)[1].data * BackLevel
        if (Back is None) and (BackLevel is None):
            # Use no background
            print("No background added")
            final = 0.
        if (type(Back)==np.ndarray) and (BackLevel is not None):
            # Use a passed background and scale its median to BackLevel
            print("adding passed background array scaled to",BackLevel)
            final = Back/np.median(Back) * BackLevel
        if (type(Back)==np.ndarray) and (BackLevel is None):
            # Use a passed background as is
            print("adding passed background array as is")
            final = Back

        if not ((Back is None) and (BackLevel is None)) and (tofits!=None):
            # Save the background image to a fits file
            hprime = fits.PrimaryHDU()
            himg = fits.ImageHDU(final)
            himg.header['EXTNAME'] = 'BACKGRND'
            himg.header['UNITS'] = 'e/s'
            if BackLevel is not None:
                himg.header['BackLevel'] = BackLevel
            hlist = fits.HDUList([hprime, himg])
            hlist.writeto(tofits, overwrite=True)

        for order in self.orders:
            print("Adding contribution from order ",order)
            try:
                sim = self.this_one[order].simulated_image[self.ystart:self.yend+1,self.xstart:self.xend+1]
            except AttributeError:
                print("Contribution from order",order,"is missing. Skipping it.")
                continue
            final = final + sim
        self.final = final  
        
        if (Back is None) and (BackLevel is None) and (tofits!=None):
            # Save the background image to a fits file
            hprime = fits.PrimaryHDU()
            himg = fits.ImageHDU(final)
            himg.header['EXTNAME'] = 'BACKGRND'
            himg.header['UNITS'] = 'e/s'
            if BackLevel is not None:
                himg.header['BackLevel'] = BackLevel
            hlist = fits.HDUList([hprime, himg])
            hlist.writeto(tofits, overwrite=True)

    def saveSingleFits(self,name):
        """A helper function to write the 2D simulated imae into a fits file

        Parameters
        ----------
        fname: str
            The file name to use for the output
        """

        #save an array into the first extension of a fits file
        h0 = fits.PrimaryHDU()
        h1 = fits.ImageHDU(self.final,name='DATA')
        
        
        hdulist = fits.HDUList([h0,h1])
        hdulist.writeto(name,overwrite=True)

if __name__ == '__main__':
    import glob

    image_seeds = glob.glob("V4*.fits")
    seed = Grism_seed(image_seeds,"F444W","modA_R","/Users/npirzkal/Dropbox/GRISMDATA/NIRCAM/")






import os
from setuptools import setup, Extension, find_packages

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
	name='NIRCAM_Gsim',
	version='1.71',
	description='NIRCAM Grism simulator, includes J.D. Smith polyclip C code',
    url='https://github.com/npirzkal/GRISM_NIRCAM',
	author='Nor Pirzkal',
	author_email='npirzkal@mac.com',
    package_dir = {
        'NIRCAM_Gsim': 'NIRCAM_Gsim',
        'NIRCAM_Gsim.polyclip': 'NIRCAM_Gsim/polyclip',
        'NIRCAM_Gsim.observations': 'NIRCAM_Gsim/observations'},
    packages=["NIRCAM_Gsim","NIRCAM_Gsim.polyclip","NIRCAM_Gsim.observations","NIRCAM_Gsim.disperse"],
    ext_modules=[Extension('NIRCAM_Gsim.polyclip.polyclip_c', ['NIRCAM_Gsim/polyclip/polyclip_c.c'])],
    install_requires=[
    "tqdm > 4.0.0",
    "grismconf >= 1.24"
],
)



from scipy.interpolate import interp1d
import numpy as np
from ..polyclip.polyclip import polyclip


def dispersed_pixel(x0s,y0s,f0,order,C,ID,oversample_factor=2,extrapolate_SED=False,xoffset=0,yoffset=0,outbound=False):
    """This function take a list of pixels and disperses them using the information contained
    in a GRISMCONF file, and returns a list of pixel and fluxes.

    Parameters
    ----------
    x0s: list
        A list of n x-coordinates.
    y0s: list
        A list of n y-coordinates.
    f0: list
        A list of n flux (flam) for each of the pixels contained in x0s,y0s.
    order: str
        The name of the spectral order to disperse.
    ID: int
        The ID of the object this is for.
    oversample_factor: int
        The amount of oversampling required above that of the input spectra or natural dispersion, whichever is smaller. Default=2.
    extrapolate_SED: bol
        Whether to allow for the SED of the object to be extrapolated when it does not fully cover the 
        needed wavelength range. Default if False.
    xoffset int
        A pixel offset to apply when computing the dispersion (for padded images for example)
    yoffset int
        A pixel offset to apply when computing the dispersion (for padded images for example)


    Output
    ------
    xs: array
        A list of x-coordinates dispersed pixel
    ys: array
        A list of y-coordinates dispersed pixel
    areas: array
        A list of the areas of the incident pixel thatm when dispersed falls on the dispersed pixel
    lams: array
        A list of the wavelength of each of the dispersed pixels
    counts: array
        A list of counts for each of the dispersed pixels
    ID: array
        A list containing the ID. Returned for bookkeeping convenience.
    """
    if len(f0[0])>1:
        #print("f0:",f0,len(f0[0]),len(f0[1]))
        if extrapolate_SED==False:
            f = interp1d(f0[0],f0[1],fill_value=0.,bounds_error=False)
        else:
            f = interp1d(f0[0],f0[1],fill_value="extrapolate",bounds_error=False)
    else:
        f = lambda x: f0[1][0]

    s = C.SENS[order]

    x0 = np.mean(x0s)
    y0 = np.mean(y0s)

    x00 = x0+xoffset
    y00 = y0+yoffset

    if outbound is False:
        if x00<0: x00=0
        if y00<0: y00=0
        if x00>C.NAXIS[0]: x00=C.NAXIS[0]
        if y00>C.NAXIS[1]: y00=C.NAXIS[1]



    dx0s = [t-x0 for t in x0s]
    dy0s = [t-y0 for t in y0s]
    
    # Figuring out a few things about size of order, dispersion and wavelengths to use    
    wmin = C.WRANGE[order][0]
    wmax = C.WRANGE[order][1]

    t0 = C.INVDISPL(order,x00,y00,wmin)
    t1 = C.INVDISPL(order,x00,y00,wmax)
    
    dx0 = C.DISPX(order,x00,y00,t0) - C.DISPX(order,x00,y00,t1)
    dx1 = C.DISPY(order,x00,y00,t0) - C.DISPY(order,x00,y00,t1)

    dw = np.abs((wmax-wmin)/(dx1-dx0))

    # Use a natural wavelength scale or the wavelength scale of the input SED/spectrum, whichever is smaller, divided by oversampling requested
    if len(f0[0])>1:
        input_dlam = np.median(f0[0][1:]-f0[0][:-1])
        if input_dlam<dw:
            dlam = input_dlam/oversample_factor
        else:
            dlam = dw/oversample_factor
    else:
        dlam = dw/oversample_factor
    

    lambdas = np.arange(wmin,wmax+dlam,dlam)

    dS = C.INVDISPL(order,x00,y00,lambdas)

    m = len(lambdas)

    dXs = C.DISPX(order,x00,y00,dS)
    dYs = C.DISPY(order,x00,y00,dS)

    x0s = x0 + dXs 
    y0s = y0 + dYs 

    padding = 1
    l = x0s.astype(np.int32) - padding
    r = x0s.astype(np.int32) + padding
    b = y0s.astype(np.int32) - padding
    t = y0s.astype(np.int32) + padding

    px = np.array([x0s+dx0s[0],x0s+dx0s[1],x0s+dx0s[2],x0s+dx0s[3]],dtype=np.float32).transpose().ravel()
    py = np.array([y0s+dy0s[0],y0s+dy0s[1],y0s+dy0s[2],y0s+dy0s[3]],dtype=np.float32).transpose().ravel()

    lams = np.array([[ll,0,0,0] for ll in lambdas],dtype=np.float32).transpose().ravel()

    poly_inds = np.arange(0,(m+1)*4,4,dtype=np.int32)

    n_poly = len(x0s)

    n = len(lams) # number of pixels we are "dropping", e.g. number of wav bins

    n = n*2
 
    index = np.zeros(n, dtype=np.int32)
    x = np.zeros(n, dtype=np.int32)
    y = np.zeros(n, dtype=np.int32)
    areas = np.zeros(n, dtype=np.float32)
    nclip_poly = np.array([0],np.int32)
    polyclip.polyclip_multi4(l,r,b,t,px,py,n_poly,poly_inds,x,y,nclip_poly,areas,index)


    xs  = x[0:nclip_poly[0]]
    ys  = y[0:nclip_poly[0]]
    areas = areas[0:nclip_poly[0]]
    lams = np.take(lambdas,index)[0:len(xs)]
    counts = f(lams)*areas*s(lams)*np.abs(dlam) * 10000. # factor of 10000 because dlam is in micron and we want Angstrom with to apply f(lams)
    vg = (xs>=0) & (ys>=0)

    if len(xs[vg])==0:
        return np.array([0]),np.array([0]),np.array([0]),np.array([0]),np.array([0]),0
    return xs[vg], ys[vg], areas[vg], lams[vg], counts[vg], ID
